/**
* Module:  Tutorial code for the XK-1
* Version: 0.1
* File:    test.xc
*
* The copyrights, all other intellectual and industrial
* property rights are retained by XMOS and/or its licensors.
* Terms and conditions covering the use of this code can
* be found in the Xmos End User License Agreement.
*
* Copyright XMOS Ltd 2009
**/

#include <xs1.h>
#include <print.h>

// LEDs
out port p_leds_3_0 = XS1_PORT_4F;


// Buttons
in port p_button_0 = XS1_PORT_1K;
in port p_button_1 = XS1_PORT_1L;


void flashLeds(out port leds, chanend c)
{  
  timer t;
  unsigned int time;
  int ledVal = 1;
  int isOn = 1;
  int val;
  t :> time;

  leds <: ledVal;
  while (1)
  {
    
    select
    {
      // flash the led on and off, with a period of 1Hz
      case t when timerafter(time) :> void:
      {
        if (isOn)
          leds <: ledVal;
        else
          leds <: 0;
        
        isOn = !isOn;
        
        time += 50000000;
        break;
      }
      
      // receive information from the button to decide which leds to flash
      case c :> ledVal :
      {
        break;
      }
    }
  }
}

void buttonListener(in port button, chanend c)
{
  int led = 1;
  while (1)
  {
    select
    {
      // wait for button to be pressed
      case button when pinseq(0) :> void:
      {
        c <: led;
        led = (led + 1) & 0xF;
        
        // wait for button to be released
        button when pinseq(1) :> void;
        
        break;
      }
    }
  }
}


int main()
{
  chan c;
  par
  {
    flashLeds(p_leds_3_0, c);
    buttonListener(p_button_0, c);
  }
  return 0;
}
